/*---------------------------------------------------------------------------------------------------------------------------------------------------
 * @file hexedit.c - full screen hex editor
 *
 * Copyright (c) 2014 Frank Meyer - frank(at)fli4l.de
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Possible keys:
 *  LEFT                back one byte
 *  RIGHT               forward one byte
 *  DOWN                one line down
 *  UP                  one line up
 *  TAB                 toggle between hex and ascii columns
 *  any other           input as hex or ascii value
 *  2 x ESCAPE          exit
 *---------------------------------------------------------------------------------------------------------------------------------------------------
 */

/*---------------------------------------------------------------------------------------------------------------------------------------------------
 * @file hexedit.c - full screen hex editor
 *
 * Copyright (c) 2014 Frank Meyer - frank(at)fli4l.de
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Possible keys:
 *  LEFT                back one byte
 *  RIGHT               forward one byte
 *  DOWN                one line down
 *  UP                  one line up
 *  TAB                 toggle between hex and ascii columns
 *  any other           input as hex or ascii value
 *  2 x ESCAPE          exit
 *---------------------------------------------------------------------------------------------------------------------------------------------------
 */

#include <stdint.h>
#include <string.h>
#include <ctype.h>
#include "mcurses.h"
#include "hexedit.h"

/*---------------------------------------------------------------------------------------------------------------------------------------------------
 * itox: convert a decimal value 0-15 into hexadecimal digit
 *---------------------------------------------------------------------------------------------------------------------------------------------------
 */
static void itox (uint8_t val)
{
    uint8_t ch;

    val &= 0x0F;

    if (val <= 9)
    {
        ch = val + '0';
    }
    else
    {
        ch = val - 10 + 'A';
    }
    addch (ch);
}

/*---------------------------------------------------------------------------------------------------------------------------------------------------
 * itoxx: convert a decimal value 0-255 into 2 hexadecimal digits
 *---------------------------------------------------------------------------------------------------------------------------------------------------
 */
static void itoxx (unsigned char i)
{
    itox (i >> 4);
    itox (i & 0x0F);
}

static uint8_t xtoi (uint8_t ch)
{
    uint8_t val;

    ch = toupper(ch);    
    if (ch >= 'A' && ch <= 'F')
        val = (ch - 'A') + 10;
    else
        val = (ch - '0');

    itox (val);     // print value in hex
    return val;
}

static void print_hex_line (uint8_t line, uint16_t off)
{
    uint8_t         col;
    uint8_t         ch;

    move (line, 0);
    itoxx (off >> 8);
    itoxx (off & 0xFF);

    move (line, FIRST_HEX_COL);
    for (col = 0; col < BYTES_PER_ROW; col++)
    {
        itoxx (PEEK(off));
        addch (' ');
        off++;
    }

    off -= BYTES_PER_ROW;

    move (line, FIRST_ASCII_COL);

    for (col = 0; col < BYTES_PER_ROW; col++)
    {
        ch = PEEK(off);

        if (IS_PRINT(ch))
        {
            addch (ch);
        }
        else
        {
            addch ('.');
        }
        off++;
    }
}

uint16_t off_start;

static void show_page (uint16_t off)
{
    uint8_t         line;

    off_start = off;

    for (line = FIRST_LINE; line < LAST_LINE; line++)
    {
        print_hex_line (line, off);
        off += BYTES_PER_ROW;
    }
}

void show_status_line ()
{
    move (LINES - 1, 0);
    attrset (F_WHITE | B_BLUE);
    addstr ("TAB: Toggle HEX/ASCII         CTRL-G: Goto Address          2 x ESC: Exit");
    attrset (A_NORMAL);
}

/*---------------------------------------------------------------------------------------------------------------------------------------------------
 * hexdit: hex editor
 *---------------------------------------------------------------------------------------------------------------------------------------------------
 */
void hexedit (uint16_t offset)
{
    uint8_t         ch;
    uint8_t         line;
    uint8_t         col;
    uint16_t        off;
    uint8_t         byte;
    uint8_t         mode = MODE_HEX;

    clear ();
    setscrreg (FIRST_LINE, LAST_LINE - 1);

    off = offset;

    move (0, 0);
    attrset (F_WHITE | B_RED);

    addstr ("Addr");

    for (col = 4; col < FIRST_HEX_COL; col++)
    {
        addch (' ');
    }

    for (byte = 0; byte < BYTES_PER_ROW; byte++)
    {
        itoxx (byte);
        addch (' ');
        col += 3;
    }

    for ( ; col < FIRST_ASCII_COL; col++)
    {
        addch (' ');
    }

    for (byte = 0; byte < BYTES_PER_ROW; byte++)
    {
        itox (byte);
    }

    attrset (A_NORMAL);

    show_page (off);

    show_status_line ();

    off = offset;
    line = FIRST_LINE;
    col = FIRST_HEX_COL;
    byte = 0;

    do
    {
        move (line, col);
        ch = getch ();

        switch (ch)
        {
            case KEY_RIGHT:
            {
                if (byte < BYTES_PER_ROW - 1)
                {
                    byte++;
                    col += (mode == MODE_HEX) ? 3 : 1;
                }
                break;
            }
            case KEY_LEFT:
            {
                if (byte > 0)
                {
                    byte--;
                    col -= (mode == MODE_HEX) ? 3 : 1;
                }
                break;
            }
            case KEY_DOWN:
            {
                if (off < 65520)
                {
                    off += BYTES_PER_ROW;

                    if (line == LAST_LINE - 1)
                    {
                        scroll ();
                        print_hex_line (line, off);
                        off_start += BYTES_PER_ROW;
                    }
                    else
                    {
                        line++;
                    }
                }
                break;
            }
            case KEY_NPAGE:
            {
                off = off_start + (LAST_LINE - FIRST_LINE) * BYTES_PER_ROW;
                show_page (off);
                line = FIRST_LINE;
                col = FIRST_HEX_COL;
                mode = MODE_HEX;
                byte = 0;
                break;
            }
            case KEY_UP:
            {
                if (off >= BYTES_PER_ROW)
                {
                    off -= BYTES_PER_ROW;

                    if (line == FIRST_LINE)
                    {
                        insertln ();
                        print_hex_line (line, off);
                        off_start -= BYTES_PER_ROW;
                    }
                    else
                    {
                        line--;
                    }
                }
                break;
            }
            case KEY_PPAGE:
            {
                off = off_start - (LAST_LINE - FIRST_LINE) * BYTES_PER_ROW;
                show_page (off);
                line = FIRST_LINE;
                col = FIRST_HEX_COL;
                mode = MODE_HEX;
                byte = 0;
                break;
            }
            case KEY_HOME:
            {
                byte = 0;
                col = (mode == MODE_HEX) ? FIRST_HEX_COL : FIRST_ASCII_COL;
                break;
            }
            case KEY_END:
            {
                byte = BYTES_PER_ROW - 1;
                col = (mode == MODE_HEX) ? (FIRST_HEX_COL + 3 * byte) : (FIRST_ASCII_COL + byte);
                break;
            }
            case KEY_TAB:
            {
                if (mode == MODE_HEX)
                {
                    mode = MODE_ASCII;
                    col = FIRST_ASCII_COL + byte;
                }
                else
                {
                    mode = MODE_HEX;
                    col = FIRST_HEX_COL + 3 * byte;
                }
                break;
            }
            case KEY_CTRL('G'):
            {
                char    buf[5];
                uint8_t i;

                move (LINES - 1, 0);
                addstr ("Goto address: ");
                clrtoeol ();
                getnstr (buf, 5);

                off = 0;

                for (i = 0; buf[i]; i++)
                {
                    if (i > 0)
                    {
                        off <<= 4;
                    }
                    off += xtoi (buf[i]);
                }

                byte = off & 0x000F;
                off &= 0xFFF0;

                show_status_line ();
                show_page (off);
                line = FIRST_LINE;

                if (mode == MODE_HEX)
                {
                    col = FIRST_HEX_COL + 3 * byte;
                }
                else
                {
                    col = FIRST_ASCII_COL + byte;
                }
                break;
            }
            default:
            {
                if (mode == MODE_HEX)
                {
                    if (IS_HEX(ch))
                    {
                        uint16_t    addr  = off + byte;
                        uint8_t     value = xtoi (ch) << 4;

                        ch = getch ();

                        if (IS_HEX(ch))
                        {
                            uint8_t value_read;

                            value |= xtoi (ch);
                            POKE(addr, value);

                            value_read = PEEK (addr);                                // read it back, perhaps it is Z80 ROM ;-)

                            if (value != value_read)
                            {
                                move (line, FIRST_HEX_COL + 3 * byte);
                                itoxx (value_read);
                            }

                            move (line, FIRST_ASCII_COL + byte);

                            if (IS_PRINT(value))
                            {
                                addch (value);
                            }
                            else
                            {
                                addch ('.');
                            }

                            if (byte < BYTES_PER_ROW - 1)   // next byte...
                            {
                                byte++;
                                col += 3;
                            }
                        }
                    }
                }
                else // MODE_ASCII
                {
                    if (IS_PRINT(ch))
                    {
                        uint16_t addr = off + byte;
                        POKE(addr, ch);
                        ch = PEEK (addr);                                       // read it back, perhaps it is Z80 ROM ;-)
                        addch (ch);
                        move (line, FIRST_HEX_COL + 3 * byte);
                        itoxx (ch);

                        if (byte < BYTES_PER_ROW - 1)   // next byte...
                        {
                            byte++;
                            col++;
                        }
                    }
                }
            }
        }
    } while (ch != KEY_ESCAPE);
}

