#include <asm/z80/features.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>
#include "main.h"
#include "mcurses.h"
#include "driver.h"
#include "command.h"
#include "tools.h"


// Game of Life

#define   VT100_MAX_X         79
#define   VT100_MAX_Y         23
#define   MIN_X               0
#define   MIN_Y               0
#define   MAX_X               79        // VT100 Terminal Max Koords
#define   MAX_Y               22
#define   MAX_ZELLEN          500       // Anzahl der zum Start bevölkerten Felder

// Zustände der nächsten Generation
#define DEAD    0
#define ALIVE   1

// 2 Spielfelder alt+neu
struct feld_t {
      uint8_t last[MAX_X+1][MAX_Y+1];
      uint8_t next[MAX_X+1][MAX_Y+1];
};

struct feld_t feld;

/*
    23/3
    Regel       Geburt:                     = 3 Nachbarn
                Stirbt Einsamkeit:          < 2 Nachbarn
                Stirbt Überbevölkerjung:    > 3 Nachbarn
                Bleibt am Leben:            2 oder 3 Nachbarn

 */

// Erzeuge Startzahl für Zufallsgenerator
uint8_t rreg;
uint8_t get_rreg()
{
    __asm
        ld a,r
        ld (_rreg),a
    __endasm;

    return(rreg);
}

// Setze ein zufälliges Startmuster
void fill_feld_random()
{
    uint16_t k;
    uint8_t x,y;

     k = MAX_ZELLEN;
     do {
         x = prng(MAX_X);
         y = prng(MAX_Y);
         feld.last[x][y] = ALIVE;
         move(MAX_Y-y,x);
         addch('O');
         k--;
     } while (k > 0);
}


// Berechne die nächste Generation
uint16_t calc_nextgen(int* cells)
{
    static int8_t x,y;              // Laufvariablen Nachbarn
    static uint8_t xpos,ypos;       // Laufvariablen Feld
    static uint8_t *foo;            // Hilfsvariable
    static uint8_t x1,y1;           // Hilfskoordinaten
    static int8_t a1,b1,a2,b2;      // Hilfsvariablen signed
    static uint16_t changes;

   // Kalkulationsfeld löschen
   memset(feld.next,0,sizeof(feld.next));
   changes = 0;

   // Durchsuche Feld nach LIVE Zellen
   for (xpos = MIN_X;xpos <= MAX_X; xpos++) {
      for (ypos = MIN_Y;ypos <= MAX_Y; ypos++) {
          if (feld.last[xpos][ypos] == ALIVE)
          {
                (*cells)++;

                // Range der 8 Nachbarn
                a1 = ypos-1;
                b1 = ypos+1;
                a2 = xpos-1;
                b2 = xpos+1;

                // Ermittle den Zustand der 8 Nachbarn
                for (y = a1;y <= b1;y++)  {
                    for(x = a2;x <= b2;x++)  {
                        // Koord.transformation auf gegenüber liegenden Seiten

                        x1 = x;
                        y1 = y;

                        if (x < 0 )             // x Koordinate
                            x1 = MAX_X;
                        else if (x > MAX_X)
                            x1 = 0;

                        if (y < 0)              // y Koordinate
                            y1 = MAX_Y;
                        else if (y > MAX_Y)
                            y1 = 0;

                        // Erhöhe Nachbarpunkte um 1
                        feld.next[x1][y1]++;
                    }
                }
                // Auf Zell Koordinate wieder 1 abziehen
                feld.next[xpos][ypos]--;
          }
      }
   }

  // -----


  // Aktualisiere das alte Feld mit Informationen
  // aus dem temporären Kalkulationsfeld
   for (x = MIN_X;x <= MAX_X; x++) {
      for (y = MIN_Y;y <= MAX_Y; y++)
      {
          foo = &feld.last[x][y];
          switch (feld.next[x][y]) {
             case 2:
                      break;
             case 3:
                      if (*foo != ALIVE)
                      {
                        *foo = ALIVE;
                        move(MAX_Y-y,x);
                        addch('O');
                        changes++;
                      }

                      break;
             default:
                      if (*foo != DEAD)
                      {
                        *foo = DEAD;
                        move(MAX_Y-y,x);
                        addch(' ');
                        changes++;
                      }
          }
      }
    }

    return (changes);

}

//  Liste sortieren
void bubble_sort(uint8_t list[], uint8_t n)
{
  uint8_t c, d, t;

  for (c = 0 ; c < ( n - 1 ); c++) {
    for (d = 0 ; d < n - c - 1; d++) {
      if (list[d] > list[d+1])  {
        /* Swapping */
        t         = list[d];
        list[d]   = list[d+1];
        list[d+1] = t;
      }
    }
  }
}

// Prüft, ob Feld nur noch aus unv. Strukturen besteht
/*
    Prinzip:
    Array wird durchsucht aus wieviel verschiedenen Zahlen
    es besteht. Bei > 4 ist die Welt noch im Fluss, bei
    < 4 befinden sich die immer gleichen Muster im Bild

*/

uint8_t check_blinker(uint16_t changes,uint8_t reset)
{
    #define MAX_HIST    16
    #define LIMIT_ZAHL  3
    #define LIMIT_MODS  250

    static uint8_t history[MAX_HIST];
    static uint8_t cnt = 0;
    static uint8_t calls = 0;
    uint8_t i,count;

    // Reset setzt alles zurück
    if (reset) {
        cnt = 0;
        calls = 0;
        memset(history,0,sizeof(history));
        LedOut(0);
        return (0);
    }

    // Erst unter N Modifikationen pro Bild suchen
    if (changes > LIMIT_MODS)
        return (0);

    // Ringpuffer füllen....
    history[cnt++] = changes;               
    cnt = cnt % MAX_HIST;

    // Die ersten Generationen ignorieren
    if (++calls < MAX_HIST)  
        return (0);

    // Liste sortieren        
    bubble_sort(history,MAX_HIST); 

    // Zahlenhäufigkeit zaehlen
    count = 0;
    for (i = 0;i < (MAX_HIST-1);i++)        
    {
        if (history[i] != history[i+1])
            count++;
    }

    LedOut(count);
    return ( (count > LIMIT_ZAHL) ? 0: 1);

}

// Hauptroutine
void life (void)
{

    uint16_t changes,generation,total_runs,cells;

    // Feld 1 löschen
    memset(feld.last,DEAD,sizeof(feld.last));

     // Meldung
    clear();
    move (11,24);
    addstr("Conways 'Game of Life'");
    delay(40000);

    // Inits
    changes    = 0;
    generation = 0;
    total_runs = 0;
    sprng(get_rreg());              // Zufall initialisieren mit R Register

    // Spiel laufen lassen
    while (1)
    {
        clear();
        curs_set(0);
        check_blinker(0,1);         //  Abbruchbedingung resetten
        fill_feld_random();         //  Feld zufaellig bespielen
        generation = 0;
        do {
            // Naechste Generation berechnen und zeichnen
            cells = 0;
            changes = calc_nextgen(&cells);
            SetDigit((uint8_t)generation % 10);
 
            move(23,20);
            clrtoeol();
            addstr("|Genr: "); addint(++generation);
            addstr(" Mods: "); addint(changes);
            addstr(" Cells: "); addint(cells);
            addstr(" Run: "); addint(total_runs);
            addstr("|");

            if (keypress()==0x0d)
                break;

        } while (!check_blinker(changes,0));
        move (11,33);
        addstr("G A M E  O V E R");
        delay(30000);
        total_runs++;
    }
}

